#define FAKESNUM 0 /* fake screen number */

#include "glx_symbols.h"

#include <dlfcn.h>
#include <stdio.h>

#include "riva_glx.h"
#include "riva_symbols.h"
#include "riva_tex.h"

/* This is where we keep per-card information, like
 * cached info on where the mapped video ram is
 * This will be allocated as riva_sinfo[numscreens]
 */
RIVA_GLX_INFO *riva_sinfo=NULL;


static void (*EnableVTSwitch)(void *);

static int any_error;
static void *my_dlsym(void *handle, const char *name)
{
   void *tmp = dlsym(handle, name);
   char *error;
   if ((error = dlerror()) != 0) {
      fputs(error, stderr);
      any_error = 1;
      return 0;
   }
   return tmp;
}


/*
 *There are a few possible ways this nasty hack can be implemented:
 * The "clean" way, with xfree doing it all.. or hacky ways.
 * Unfortunately, there is a bug or two in xfree vidmem management.
 * Sooo... hack away.
 *
 * Currently, what it does is look for unmanaged video memory that
 * is free because the user under-specified VideoRam in XF86Config.
 * We need around 8 megabytes free!! But we dont check that here.
 *
 * return true (1) on okay, false (0) on fail
 * riva_sinfo[].privMemBase,privMemSize   will be set here on okay.
 */
static int nvMapPrivMem(ScrnInfoPtr scrninfoP){
	int screennum=scrninfoP->scrnIndex;
	unsigned int byteoffset;
	char *vidphysaddr;
	int totalmem,freememsize;
	char *mappedaddr;

	/* Take total RAM size, subtract what xfree is using, and
	 * thats what we have left to play with.
	 */
	byteoffset=scrninfoP->videoRam * 1024;
	totalmem=(riva_sinfo[screennum].xfree_nvrec->riva.RamAmountKBytes)*1024;
	freememsize=totalmem-byteoffset;

	if(freememsize<2048000){
		fprintf(stderr,
			"No free video memory. You must adjust XF86Config\n");
			return 0;
	}
		
	vidphysaddr=(char*)riva_sinfo[screennum].xfree_nvrec->FbAddress;

	/* I "happen to know" that the solaris implementation of
	 * xf86MapVidMem() does not do boundary checks against the
	 * value in XF86Config. It seems that other OS implementations
	 * are similarly permissive.
	 *That makes life a lot easier for us here.
	 */
	mappedaddr=xf86MapVidMem(screennum,VIDMEM_FRAMEBUFFER,
				 ((unsigned int)vidphysaddr)+byteoffset,
				 freememsize);
	if(mappedaddr==NULL){
		ErrorF("Could not mmap extra mem for GLX purposes");
		return 0;
	}
	fprintf(stderr,
	        "DEBUG: nvMapPrivMem: mapped %p(voffset 0x%x) to mem 0x%p-0x%p\n",
		vidphysaddr, byteoffset, mappedaddr,mappedaddr+freememsize-1);

	riva_sinfo[screennum].privMemBase=mappedaddr;
	riva_sinfo[screennum].privMemOffset=byteoffset;
	riva_sinfo[screennum].privMemSize=freememsize;

	return 1;
}

/* 
 * Very important routine.
 * Not only is this routine responsible for "Hooking in XFree server symbols"
 * but also for a lot of large memory allocation:
 *    BACK BUFFER
 *    DEPTH BUFFER
 *    TEXTURE [memory, not exactly a "buffer"]
 * I would LIKE to do this in the more appropriately named nvInitGLX,
 * however, that routine depends on this stuff already being set up.
 *
 * return true (1) on okay, false (0) on fail
 */
int nvHookServerSymbols( void *handle )
{
   int snum=FAKESNUM; /* XXX hardcode screennum. Bad Boy */
   ScrnInfoPtr scrninfoP;
   ScreenPtr screenP;
   NVPtr nvptr;
   int bytesperpixel, dbuffersize,depthbufsize, texturebufsize;
   char *currfreeptr;
   int freeoffset, freebytesleft;

   fprintf(stderr,"DEBUG: in nvHookServerSymbols\n");
   any_error = 0;
   

   /* Hook into symbols that are in the XFree nv driver specifically */
   if(riva_sinfo !=NULL){
	ErrorF("nvHookServerSymbols: riva_sinfo not NULL?\n");
	return 0;
   }
   scrninfoP=glxsym.scrninfoList[snum];
   screenP=screenInfo.screens[scrninfoP->scrnIndex];

   nvptr=(NVPtr) scrninfoP->driverPrivate;

   riva_sinfo=(RIVA_GLX_INFO*)xalloc(sizeof (RIVA_GLX_INFO) * screenInfo.numScreens);
   if(!riva_sinfo){
	ErrorF("nvHookServerSymbols: xalloc failed\n");
	return 0;
   }
   bzero(riva_sinfo,sizeof (RIVA_GLX_INFO) * screenInfo.numScreens);

   bytesperpixel=(scrninfoP->bitsPerPixel) >>3;

   if(scrninfoP->bitsPerPixel & 0xf){
	/* 15 bits = 1 + 1 bytes*/
	bytesperpixel+=1;
   }
   fprintf(stderr,"DEBUG: nvHookServerSym: using %d bytesperpixel\n",
	   bytesperpixel);

   riva_sinfo[snum].xfree_nvrec= nvptr;
   riva_sinfo[snum].riva=&nvptr->riva;

   if(!nvMapPrivMem(scrninfoP)){
	ErrorF("nvHookServerSymbols bailing out\n");
	return 0;
   }

   currfreeptr=riva_sinfo[snum].privMemBase;
   freebytesleft=riva_sinfo[snum].privMemSize;
   freeoffset=riva_sinfo[snum].privMemOffset; /*offset from vidram 0*/

   riva_sinfo[snum].videoareasize=bytesperpixel*
				scrninfoP->virtualX * scrninfoP->virtualY;
   dbuffersize=riva_sinfo[snum].videoareasize;

   fprintf(stderr,"DEBUG: nvHookServerSym: initial free bytes=%d\n",
	   freebytesleft);

   /************************************************************************
    * Now starts all the big backbuffer, etc...  allocation.               *
    *                                                                      *
    * We are more likely to need a "back buffer" for doublebuffering,      *
    * then a "depth buffer". But try for both.                             *
    * Due to nastiness in xfree memory management routines, we have done   *
    * our own reserving of some videoram space. We now likewise do our own *
    * "memory allocation" from a superchunk that we have hanging off       *
    * currfreeptr. SIGH.
    ***********************************************************************/

   /* This is just assumed to always be 0, but it's a consistency thing
    * to have it defined .
    */
   riva_sinfo[snum].frontbufferoffset=0;


   /* Allocate "Depth Buffer", for z-buffering */
   /* It seems the TNT hardware is always 16bits for depth buffers.
    * See RivaGetDepth(). If this changes, will have to
    * initialize better here
    */
   riva_sinfo[snum].depth_bits=16;

   depthbufsize=(riva_sinfo[snum].depth_bits>>3) *
		scrninfoP->virtualX * scrninfoP->virtualY;

   if(freebytesleft<depthbufsize){
	ErrorF("nvHookServerSymbols: alloc of depth buffer failed: %d free, need %d\n",
	       freebytesleft, depthbufsize);
	return 0;
   }

   riva_sinfo[snum].depthbufferoffset=freeoffset;
   riva_sinfo[snum].depthbufmappedaddr=currfreeptr;
   freeoffset +=depthbufsize; currfreeptr +=depthbufsize;
   freebytesleft -=depthbufsize;

   fprintf(stderr,"DEBUG: nvHookServerSym: after depthbuf, free bytes=%d\n",
	   freebytesleft);

   /*
    * Allocate "Back Buffer", for double--buffering
    */
   if(freebytesleft<dbuffersize){
	ErrorF("nvHookServerSymbols: alloc of back buffer failed: %d free, need %d\n",
	       freebytesleft, dbuffersize);
	return 0;
   }

   riva_sinfo[snum].backbufferoffset=freeoffset;
   riva_sinfo[snum].backbufmappedaddr=currfreeptr;
   freeoffset +=dbuffersize; currfreeptr +=dbuffersize;
   freebytesleft -=dbuffersize;

   fprintf(stderr,"DEBUG: nvHookServerSym: after doublebuff, free bytes=%d\n",
	   freebytesleft);

   /*
    * Allocate remaining memory as "Texture (Area)", for
    * loading textures onto the card 
    * XXX in future,  we should be using AGP instead, somehow.
    */

   riva_sinfo[snum].numtexblocks=freebytesleft / RIVA_TEX_BLOCK_SIZE;
   fprintf(stderr,"HACK: decreasing number of tex blocks allocated manually\n");
   riva_sinfo[snum].numtexblocks-=15; /* XXX Nasty blind hack */
                                      /* but it WORKS, dammit.
				       * Guess I need to reserve some memory
				       * at the "end" of videomem.
				       * Be nice to find out exactly how much.
				       */

   if(riva_sinfo[snum].numtexblocks < NVGLX_NUMTEXTUREBLOCKS){
	ErrorF("nvHookServerSymbols: alloc of texture mem failed\n");
	return 0;
   }
   texturebufsize=riva_sinfo[snum].numtexblocks * RIVA_TEX_BLOCK_SIZE;
   freebytesleft -= texturebufsize;

   fprintf(stderr,"DEBUG: will alloc %d texture blocks(0x%x/%d bytes) at %p\n",
	   riva_sinfo[snum].numtexblocks, texturebufsize, texturebufsize,
	   currfreeptr);

    riva_sinfo[snum].textureoffset=freeoffset;
    riva_sinfo[snum].texturebase=currfreeptr;



   ErrorF("   GLX disabling VT switching\n");
    
   EnableVTSwitch=my_dlsym(handle,"xf86EnableVTSwitch");

   if(EnableVTSwitch){
	ErrorF("Disabling VT switching, for initial GLX version\n");
	(*EnableVTSwitch)(FALSE);
   }

   if (!any_error) 
      ErrorF("Sucessfully linked in GLX NV subdriver\n");
   else
      ErrorF("GLX NV driver disabled due to missing symbols\n");

   fprintf(stderr,"DEBUG: nvHook wrapup: depthoffset=%x, backoffset=%x, textureoffset=%lx\n",
		riva_sinfo[0].depthbufferoffset,
		riva_sinfo[0].backbufferoffset,
		riva_sinfo[0].textureoffset);

   return !any_error;
}
